import warnings
import numpy as np
import pandas as pd
import itertools
import random as rn
import argparse
import time
import os
import sys
#from XGB_model import tvt_xgb
#from SVM_model import tvt_svm
#from RF_model import tvt_rf
#from KNN_model import tvt_knn
#from DNN_model import tvt_dnn
from graph_model import tvt_dl
np.set_printoptions(threshold=sys.maxsize)
#Random seeding
os.environ['PYTHONHASHSEED']='0'
np.random.seed(123)
rn.seed(123)
warnings.filterwarnings('ignore')
# loading data
def model_set(X,Y,split_type='random',FP_type='ECFP4',model_type = 'SVM',model_dir=False,file_name=None,device = 'cpu',difftasks='activity'):

    if model_type == 'gcn':
        # print(difftasks)
        par, bes = tvt_dl(X, Y, split_type=split_type,file_name=file_name,model_name='gcn',model_dir=model_dir,device =device,difftasks=difftasks)
        return par, bes
    elif model_type == 'attentivefp':
        par, bes = tvt_dl(X, Y, split_type=split_type, file_name=file_name, model_name='attentivefp', model_dir=model_dir,device =device,difftasks=difftasks)
        return par,bes
    elif model_type == 'gat':
        par, bes = tvt_dl(X, Y, split_type=split_type, file_name=file_name, model_name='gat',
                            model_dir=model_dir,device =device,difftasks=difftasks)
        return par,bes
    elif model_type == 'mpnn':
        par, bes = tvt_dl(X, Y, split_type=split_type, file_name=file_name, model_name='mpnn',
                            model_dir=model_dir,device =device,difftasks=difftasks)
        return par,bes

def pair_param(file_name,data,split_type,model_type,device,difftasks):#FP_type,
    X = data.cano_smiles
    Y = data['NR-AR']
    model_path = 'model_save/'+model_type
    model_dir = file_name.replace(file_name.split('/')[-1],model_path)
    result_path = 'result_save/' + model_type
    result_dir = file_name.replace(file_name.split('/')[-1], result_path)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    if not os.path.exists(result_dir):
        os.makedirs(result_dir)
    para_name, best_name = os.path.join(result_dir, '_'.join([split_type, model_type, 'para.csv'])), os.path.join(
        result_dir, '_'.join([split_type, model_type, 'best.csv']))
    if os.path.exists(para_name):
        print(para_name,"exists files")
        pass
    else:
        # print(difftasks)
        para,best = model_set(X,Y, split_type=split_type,model_type=model_type,model_dir=model_dir,file_name=file_name,device = device,difftasks=difftasks)
        para.to_csv(para_name,index=False)
        best.to_csv(best_name,index=False)
        pass
    return

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--file', required=True, help='we must give this para')
    parser.add_argument('--split', default=['scaffold'], nargs='*',choices = ['random','scaffold','cluster'])
    parser.add_argument('--model', default=['gcn'], nargs='*',choices = ['gcn','gat','mpnn','attentivefp'])
    parser.add_argument('--difftasks', default='activity', choices=['activity', 'tox21'])
    parser.add_argument('--device', default='cpu', choices=['cpu', 'gpu'])
    args = parser.parse_args()
    return args
args = parse_args()
file_name = args.file
split = args.split
model = args.model
difftasks = args.difftasks
data = pd.read_csv(file_name,error_bad_lines=False)
device = args.device
dic = {'split':split,'model':model}
hy = pd.DataFrame(list(itertools.product(*dic.values())),columns=dic.keys())

for i in range(len(hy.index.values)):
    split_type = hy.iloc[i].split
    model_type = hy.iloc[i].model
    start_time = time.time()
    a = pair_param(file_name,data,split_type,model_type,device,difftasks)
    use_time = int(time.time() - start_time)
    if use_time >= 10:
        local_time = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime())
        f = open('../dataset/record_every_model.csv', 'a+')
        f.write(",".join([file_name, split_type,model_type, str(use_time),local_time,device,"\n"]))
        f.close()



import torch
import numpy as np
import pandas as pd
from dnn_torch_utils import Meter, MyDataset, EarlyStopping, MyDNN, collate_fn, set_random_seed
from hyperopt import fmin, tpe, hp, rand, STATUS_OK, Trials, partial
import sys
import copy
from torch.utils.data import DataLoader
from torch.nn import BCEWithLogitsLoss, MSELoss
import gc
import time
start_time = time.time()
from sklearn.model_selection import train_test_split
import warnings
from splitdater import split_dataset
from feature_create  import create_des
warnings.filterwarnings('ignore')
torch.backends.cudnn.enabled = True
torch.backends.cudnn.benchmark = True
set_random_seed(seed=43)


def run_a_train_epoch(model, data_loader, loss_func, optimizer, args):
    model.train()

    train_metric = Meter()  # for each epoch
    for batch_id, batch_data in enumerate(data_loader):
        Xs, Ys, masks = batch_data

        # transfer the data to device(cpu or cuda)
        Xs, Ys, masks = Xs.to(args['device']), Ys.to(args['device']), masks.to(args['device'])

        outputs = model(Xs)
        loss = (loss_func(outputs, Ys) * (masks != 0).float()).mean()

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        outputs.cpu()
        Ys.cpu()
        masks.cpu()
        loss.cpu()
#        torch.cuda.empty_cache()

        train_metric.update(outputs, Ys, masks)
    if args['reg']:
        rmse_score = np.mean(train_metric.compute_metric(args['metric']))  # in case of multi-tasks
        mae_score = np.mean(train_metric.compute_metric('mae'))  # in case of multi-tasks
        r2_score = np.mean(train_metric.compute_metric('r2'))  # in case of multi-tasks
        return {'rmse': rmse_score, 'mae': mae_score, 'r2': r2_score}
    else:
        roc_score = np.mean(train_metric.compute_metric(args['metric']))  # in case of multi-tasks
        prc_score = np.mean(train_metric.compute_metric('prc_auc'))  # in case of multi-tasks
        return {'roc_auc': roc_score, 'prc_auc': prc_score}


def run_an_eval_epoch(model, data_loader, args):
    model.eval()

    eval_metric = Meter()
    with torch.no_grad():
        for batch_id, batch_data in enumerate(data_loader):
            Xs, Ys, masks = batch_data
            # transfer the data to device(cpu or cuda)
            Xs, Ys, masks = Xs.to(args['device']), Ys.to(args['device']), masks.to(args['device'])

            outputs = model(Xs)

            outputs.cpu()
            Ys.cpu()
            masks.cpu()
#            torch.cuda.empty_cache()
            eval_metric.update(outputs, Ys, masks)
    if args['reg']:
        rmse_score = np.mean(eval_metric.compute_metric(args['metric']))  # in case of multi-tasks
        mae_score = np.mean(eval_metric.compute_metric('mae'))  # in case of multi-tasks
        r2_score = np.mean(eval_metric.compute_metric('r2'))  # in case of multi-tasks
        return {'rmse': rmse_score, 'mae': mae_score, 'r2': r2_score}
    else:
        roc_score = np.mean(eval_metric.compute_metric(args['metric']))  # in case of multi-tasks
        prc_score = np.mean(eval_metric.compute_metric('prc_auc'))  # in case of multi-tasks
        stats = eval_metric.compute_metric('statistical')
        return {'roc_auc': roc_score, 'prc_auc': prc_score,'statistical':stats}


def get_pos_weight(Ys):
    Ys = torch.tensor(np.nan_to_num(Ys), dtype=torch.float32)
    num_pos = torch.sum(Ys, dim=0)
    num_indices = torch.tensor(len(Ys))
    return (num_indices - num_pos) / num_pos


def standardize(col):
    return (col - np.mean(col)) / np.std(col)


def all_one_zeros(series):
    if (len(series.dropna().unique()) == 2):
        flag = False
    else:
        flag = True
    return flag

# file_name = '/home/ubuntu/Downloads/total_data.csv'  # './dataset/bace_moe_pubsubfp.csv'
# dataset_all = pd.read_csv(file_name)
# split_type = 'random'
# FP_type= 'ECFP4'
# dataset_all = dataset_all[['id', 'canoincal', 'RBA']]
# df = dataset_all
# df['RBA'] = df['RBA'].astype('int')
# df['Activitys'] = df['RBA'].apply(lambda x: 0 if x < 95 else 1)
# X = df['canoincal']
# Y = df['Activitys']


def best_model_runing(seed, opt_res, X, Y, split_type='random', FP_type='ECFP4'):
    # construct the model based on the optimal hyper-parameters
    hidden_unit1_ls = [64, 128, 256, 512]
    hidden_unit2_ls = [64, 128, 256, 512]
    hidden_unit3_ls = [64, 128, 256, 512]
    opt_hidden_units = [hidden_unit1_ls[opt_res['hidden_unit1']], hidden_unit2_ls[opt_res['hidden_unit2']],
                        hidden_unit3_ls[opt_res['hidden_unit3']]]
    # 50 repetitions based on the best model

    pd_res = []
    epochs = 300  # training epoch

    batch_size = 128
    patience = 50
    # device = torch.device('cuda:1' if torch.cuda.is_available() else 'cpu')
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # os.environ["CUDA_VISIBLE_DEVICES"] = gpu_id
    # if device == 'cuda':
    #    torch.cuda.set_device(eval(gpu_id))  # gpu device id
    reg = False
    args = {'device': device, 'metric': 'rmse' if reg else 'roc_auc', 'epochs': epochs,
            'patience': patience, 'task': '', 'reg': reg}

    data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y, split_type=split_type,
                                                                                     valid_need=True, random_state=seed)
    data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type)
    data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type)
    data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type)
    data_tr_y = data_tr_y.values.reshape(-1, 1)
    data_va_y = data_va_y.values.reshape(-1, 1)
    data_te_y = data_te_y.values.reshape(-1, 1)

    # else:
    #     training_data, data_te = train_test_split(dataset, test_size=0.1, random_state=split)
    #     # the training set was further splited into the training set and validation set
    #     data_tr, data_va = train_test_split(training_data, test_size=0.1, random_state=split)
    # prepare data for training
    # training set

    # dataloader
    train_dataset = MyDataset(data_tr_x, data_tr_y)
    validation_dataset = MyDataset(data_va_x, data_va_y)
    test_dataset = MyDataset(data_te_x, data_te_y)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    validation_loader = DataLoader(validation_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    inputs = data_tr_x.shape[1]
    best_model = MyDNN(inputs=inputs, hideen_units=opt_hidden_units, outputs=1,
                       dp_ratio=opt_res['dropout'], reg=reg)

    best_optimizer = torch.optim.Adadelta(best_model.parameters(), weight_decay=opt_res['l2'])
    file_name = './model_save/%s_%s_%s_%.4f_%d_%d_%d_%.4f_early_stop_%d.pth' % (
    FP_type, split_type, args['task'], opt_res['dropout'],
    hidden_unit1_ls[opt_res['hidden_unit1']],
    hidden_unit1_ls[opt_res['hidden_unit2']],
    hidden_unit1_ls[opt_res['hidden_unit3']],
    opt_res['l2'], seed)
    if reg:
        loss_func = MSELoss(reduction='none')
        stopper = EarlyStopping(mode='lower', patience=patience, filename=file_name)
    else:
        pos_weights = get_pos_weight(Y.values)
        loss_func = BCEWithLogitsLoss(reduction='none', pos_weight=pos_weights.to(args['device']))
        stopper = EarlyStopping(mode='higher', patience=patience, filename=file_name)
    best_model.to(device)

    for j in range(epochs):
        # training
        st = time.time()
        run_a_train_epoch(best_model, train_loader, loss_func, best_optimizer, args)
        end = time.time()
        # early stopping
        train_scores = run_an_eval_epoch(best_model, train_loader, args)
        val_scores = run_an_eval_epoch(best_model, validation_loader, args)
        early_stop = stopper.step(val_scores[args['metric']], best_model)
        if early_stop:
            break
        print(
            'task:{} repetition {:d}/{:d} epoch {:d}/{:d}, training {} {:.3f}, validation {} {:.3f}, time:{:.3f}S'.format(
                args['task'], seed, 3, j + 1, epochs, args['metric'], train_scores[args['metric']],
                args['metric'],
                val_scores[args['metric']], end - st))
        stopper.load_checkpoint(best_model)
    tr_scores = run_an_eval_epoch(best_model, train_loader, args)

    tr_results = [seed,FP_type, split_type, 'tr',tr_scores['statistical'][0], tr_scores['statistical'][1],
                  tr_scores['statistical'][2],
                  tr_scores['statistical'][2] / tr_scores['statistical'][1],
                  tr_scores['statistical'][4], tr_scores['statistical'][5],
                  tr_scores['statistical'][6], tr_scores['statistical'][7], tr_scores['statistical'][8],
                  tr_scores['statistical'][9],
                  tr_scores['statistical'][10], tr_scores['statistical'][11], tr_scores['prc_auc'],
                  tr_scores['roc_auc']]

    val_scores = run_an_eval_epoch(best_model, validation_loader, args)
    va_results = [ seed, FP_type, split_type, 'tr', val_scores['statistical'][0], val_scores['statistical'][1],
                  val_scores['statistical'][2],
                  val_scores['statistical'][2] / val_scores['statistical'][1],
                  val_scores['statistical'][4], val_scores['statistical'][5],
                  val_scores['statistical'][6], val_scores['statistical'][7], val_scores['statistical'][8],
                  val_scores['statistical'][9],
                  val_scores['statistical'][10], val_scores['statistical'][11], val_scores['prc_auc'],
                  val_scores['roc_auc']]
    te_scores = run_an_eval_epoch(best_model, test_loader, args)
    te_results = [ seed, FP_type, split_type, 'tr', te_scores['statistical'][0], te_scores['statistical'][1],
                  te_scores['statistical'][2],
                  te_scores['statistical'][2] / te_scores['statistical'][1],
                  te_scores['statistical'][4], te_scores['statistical'][5],
                  te_scores['statistical'][6], te_scores['statistical'][7], te_scores['statistical'][8],
                  te_scores['statistical'][9],
                  te_scores['statistical'][10], te_scores['statistical'][11], te_scores['prc_auc'],
                  te_scores['roc_auc']]
    pd_res.append(tr_results);pd_res.append(va_results), pd_res.append(te_results)



    return pd_res

def tvt_dnn(X,Y,split_type='random',FP_type='ECFP4'):
    hyper_paras_space = {'l2': hp.uniform('l2', 0, 0.01),
                         'dropout': hp.uniform('dropout', 0, 0.5),
                         'hidden_unit1': hp.choice('hidden_unit1', [64, 128, 256, 512]),
                         'hidden_unit2': hp.choice('hidden_unit2', [64, 128, 256, 512]),
                         'hidden_unit3': hp.choice('hidden_unit3', [64, 128, 256, 512])}

    task_type = 'cla'  # 'cla' or 'reg'
    # file_name = 'clintox_moe_pubsubfpc.csv'  # 'bace_moe_pubsubfp.csv'
    # task_type = 'cla'  # 'cla' or 'reg'
    reg = True if task_type == 'reg' else False
    epochs = 300  # training epoch

    batch_size = 128
    patience = 50
    opt_iters = 50

    # device = torch.device('cuda:1' if torch.cuda.is_available() else 'cpu')
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # os.environ["CUDA_VISIBLE_DEVICES"] = gpu_id
    # if device == 'cuda':
    #    torch.cuda.set_device(eval(gpu_id))  # gpu device id
    args = {'device': device, 'metric': 'rmse' if reg else 'roc_auc', 'epochs': epochs,
            'patience': patience, 'task': '', 'reg': reg}

    # preprocess data



    data_tr_x, data_va_x, data_te_x, data_tr_y, data_va_y, data_te_y = split_dataset(X, Y, split_type=split_type,
    valid_need=True)

    tasks = '1'

    data_tr_x, data_tr_y = create_des(data_tr_x, data_tr_y, FP_type=FP_type)
    data_va_x, data_va_y = create_des(data_va_x, data_va_y, FP_type=FP_type)
    data_te_x, data_te_y = create_des(data_te_x, data_te_y, FP_type=FP_type)
    data_tr_y = data_tr_y.values.reshape(-1,1 )
    data_va_y = data_va_y.values.reshape(-1,1 )
    data_te_y = data_te_y.values.reshape(-1,1 )
    print(len(data_tr_y),data_va_y,len(data_te_y))
    # dataloader
    train_dataset = MyDataset(data_tr_x, data_tr_y)
    validation_dataset = MyDataset(data_va_x, data_va_y)
    test_dataset = MyDataset(data_te_x, data_te_y)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    validation_loader = DataLoader(validation_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    inputs = data_tr_x.shape[1]
    if not reg:
        pos_weights = get_pos_weight(Y.values)


    def hyper_opt(hyper_paras):
        hidden_units = [hyper_paras['hidden_unit1'], hyper_paras['hidden_unit2'], hyper_paras['hidden_unit3']]
        my_model = MyDNN(inputs=inputs, hideen_units=hidden_units, dp_ratio=hyper_paras['dropout'],
                         outputs=1, reg=reg)
        optimizer = torch.optim.Adadelta(my_model.parameters(), weight_decay=hyper_paras['l2'])
        file_name = './model_save/%s_%s_%s_%.4f_%d_%d_%d_%.4f_early_stop.pth' % (FP_type,split_type,args['task'], hyper_paras['dropout'],
                                                              hyper_paras['hidden_unit1'],
                                                              hyper_paras['hidden_unit2'],
                                                              hyper_paras['hidden_unit3'],
                                                              hyper_paras['l2'])
        if reg:
            loss_func = MSELoss(reduction='none')
            stopper = EarlyStopping(mode='lower', patience=patience, filename=file_name)
        else:
            loss_func = BCEWithLogitsLoss(reduction='none', pos_weight=pos_weights.to(args['device']))
            stopper = EarlyStopping(mode='higher', patience=patience, filename=file_name)
        my_model.to(device)
        for i in range(epochs):
            # training
            run_a_train_epoch(my_model, train_loader, loss_func, optimizer, args)

            # early stopping
            val_scores = run_an_eval_epoch(my_model, validation_loader, args)

            early_stop = stopper.step(val_scores[args['metric']], my_model)

            if early_stop:
                break
        stopper.load_checkpoint(my_model)
        val_scores = run_an_eval_epoch(my_model, validation_loader, args)
        feedback = val_scores[args['metric']] if reg else (1 - val_scores[args['metric']])

        my_model.cpu()
    #    torch.cuda.empty_cache()
        gc.collect()
        return feedback


    # start hyper-parameters optimization
    trials = Trials()  # 通过Trials捕获信息
    print('******hyper-parameter optimization is starting now******')
    opt_res = fmin(hyper_opt, hyper_paras_space, algo=tpe.suggest, max_evals=opt_iters, trials=trials)

    # hyper-parameters optimization is over
    print('******hyper-parameter optimization is over******')
    print('the best hyper-parameters settings for ' + args['task'] + ' are:  ', opt_res)

    # construct the model based on the optimal hyper-parameters
    hidden_unit1_ls = [64, 128, 256, 512]
    hidden_unit2_ls = [64, 128, 256, 512]
    hidden_unit3_ls = [64, 128, 256, 512]
    opt_hidden_units = [hidden_unit1_ls[opt_res['hidden_unit1']], hidden_unit2_ls[opt_res['hidden_unit2']],
                        hidden_unit3_ls[opt_res['hidden_unit3']]]
    best_model = MyDNN(inputs=inputs, hideen_units=opt_hidden_units, outputs=1,
                       dp_ratio=opt_res['dropout'], reg=reg)
    best_file_name = './model_save/%s_%s_%s_%.4f_%d_%d_%d_%.4f_early_stop.pth' % (FP_type,split_type,args['task'], opt_res['dropout'],
                                                               hidden_unit1_ls[opt_res['hidden_unit1']],
                                                               hidden_unit1_ls[opt_res['hidden_unit2']],
                                                               hidden_unit1_ls[opt_res['hidden_unit3']],
                                                               opt_res['l2'])

    best_model.load_state_dict(torch.load(best_file_name, map_location=device)['model_state_dict'])
    best_model.to(device)
    tr_scores = run_an_eval_epoch(best_model, train_loader, args)
    val_scores = run_an_eval_epoch(best_model, validation_loader, args)
    te_scores = run_an_eval_epoch(best_model, test_loader, args)
    pd_res = []
    tr_results = [FP_type, split_type, 'tr',  tr_scores['statistical'][0], tr_scores['statistical'][1],
                              tr_scores['statistical'][2],
                              tr_scores['statistical'][2]/tr_scores['statistical'][1],
                              opt_hidden_units,[opt_res['l2'],opt_res['dropout']],tr_scores['statistical'][4],tr_scores['statistical'][5],
                  tr_scores['statistical'][6],tr_scores['statistical'][7],tr_scores['statistical'][8],tr_scores['statistical'][9],
                  tr_scores['statistical'][10],tr_scores['statistical'][11],tr_scores['prc_auc'],tr_scores['roc_auc']]
    va_results = [FP_type, split_type, 'va',  val_scores['statistical'][0], val_scores['statistical'][1],
                              val_scores['statistical'][2],
                              val_scores['statistical'][2]/val_scores['statistical'][1],
                              opt_hidden_units,[opt_res['l2'],opt_res['dropout']],val_scores['statistical'][4],val_scores['statistical'][5],
                  val_scores['statistical'][6],val_scores['statistical'][7],val_scores['statistical'][8],val_scores['statistical'][9],
                  val_scores['statistical'][10],val_scores['statistical'][11],val_scores['prc_auc'],val_scores['roc_auc']]
    te_results = [FP_type, split_type, 'te',  te_scores['statistical'][0], te_scores['statistical'][1],
                              te_scores['statistical'][2],
                              te_scores['statistical'][2]/te_scores['statistical'][1],
                              opt_hidden_units,[opt_res['l2'],opt_res['dropout']],te_scores['statistical'][4],te_scores['statistical'][5],
                  te_scores['statistical'][6],te_scores['statistical'][7],te_scores['statistical'][8],te_scores['statistical'][9],
                  te_scores['statistical'][10],te_scores['statistical'][11],te_scores['prc_auc'],te_scores['roc_auc']]
    pd_res.append(tr_results)
    pd_res.append(va_results)
    pd_res.append(te_results)
    para_res = pd.DataFrame(pd_res, columns=['FP_type', 'split_type', 'type','num_of_compounds',
                'postives',
                'negtives', 'negtives/postives',
                'para1', 'para2',
                'tn', 'fp', 'fn', 'tp', 'se', 'sp',
                'acc', 'mcc', 'auc_prc', 'auc_roc'])
    # para_res.to_csv('dnn_first.csv',index=False)
    print('training set:', tr_scores)
    print('validation set:', val_scores)
    print('test set:', te_scores)

    pd_res = []
    for i in range(3):  # pro 3
        item = best_model_runing((i + 1) * 500,opt_res, X, Y, split_type=split_type, FP_type=FP_type)
        pd_res.extend(item)
    best_res = pd.DataFrame(pd_res, columns=['seed', 'FP_type', 'split_type', 'type',
                                             'num_of_compounds', 'postives',
                                             'negtives', 'negtives/postives',
                                             'tn', 'fp', 'fn', 'tp', 'se', 'sp',
                                             'acc', 'mcc', 'auc_prc', 'auc_roc'])


    return para_res,best_res


# para,beat = tvt_dnn(X,Y,split_type='random',FP_type='ECFP4')
# print(para,beat)




import warnings
import numpy as np
import pandas as pd
import argparse
import random as rn
import multiprocessing as mp
import time
import os
import sys
from XGB_model import tvt_xgb
from SVM_model import tvt_svm
from RF_model import tvt_rf
from KNN_model import tvt_knn
from DNN_model import tvt_dnn
import itertools
np.set_printoptions(threshold=sys.maxsize)

#Random seeding
os.environ['PYTHONHASHSEED']='0'
np.random.seed(123)
rn.seed(123)
warnings.filterwarnings('ignore')

def model_set(X,Y,split_type='random',FP_type='ECFP4',model_type = 'SVM',model_dir=False,file_name=None,task_type='cla',tasks=''):
    if model_type =='SVM':
        para,best = tvt_svm(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type =='RF':
        para,best = tvt_rf(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type =='KNN':
        para,best = tvt_knn(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type == 'XGB':
        para, best = tvt_xgb(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)
    elif model_type == 'DNN':
        para, best = tvt_dnn(X, Y, split_type=split_type, FP_type=FP_type,model_dir=model_dir)

    # elif model_type == 'gcn':
    #     para, best = tvt_dl(X, Y, split_type=split_type,file_name=file_name,model_name='gcn',model_dir=model_dir)
    # elif model_type == 'attentivefp':
    #     para, best = tvt_dl(X, Y, split_type=split_type, file_name=file_name, model_name='attentivefp', model_dir=model_dir)
    # elif model_type == 'gat':
    #     para, best = tvt_dl(X, Y, split_type=split_type, file_name=file_name, model_name='gat',
    #                         model_dir=model_dir)
    # elif model_type == 'mpnn':
    #     para, best = tvt_dl(X, Y, split_type=split_type, file_name=file_name, model_name='mpnn',
    #                         model_dir=model_dir)
    return para,best

def pair_param(file_name,data,split_type,model_type,FP_type,difftasks):#,

    if len(difftasks) == 1:
        X = data.Smiles
        Y = data.activity
        model_path = 'model_save/'+model_type
        model_dir = file_name.replace(file_name.split('/')[-1],model_path)

        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        result_path = 'results_save/' + model_type
        result_dir = file_name.replace(file_name.split('/')[-1], result_path)
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)
        para_name, best_name = os.path.join(result_dir,
                                                '_'.join([split_type, model_type, FP_type, 'para.csv'])), os.path.join(
                result_dir, '_'.join([split_type, model_type, FP_type, 'best.csv']))
        if os.path.exists(para_name):
            print(para_name,'has done')
            pass
        else:
            para,best = model_set(X,Y, split_type=split_type,model_type=model_type,model_dir=model_dir,file_name=file_name,FP_type=FP_type)#
            print(best_name)
            para.to_csv(para_name,index=False)
            best.to_csv(best_name,index=False)
    else:
        for task in difftasks:
            rm_index = data[task][data[task].isnull()].index
            data.drop(index=rm_index, inplace=True)
            Y = data[task]
            X = data.cano_smiles
            print(task,split_type)
            model_path = 'model_save/'+task +'/'+ model_type
            model_dir = file_name.replace(file_name.split('/')[-1], model_path)
            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
            result_path = 'results_save/' + task + '/' + model_type
            result_dir = file_name.replace(file_name.split('/')[-1], result_path)
            if not os.path.exists(result_dir):
                os.makedirs(result_dir)
            para_name, best_name = os.path.join(result_dir,
                                                    '_'.join([split_type, model_type, FP_type, 'para.csv'])), os.path.join(
                result_dir, '_'.join([split_type, model_type, FP_type, 'best.csv']))
            print(para_name)
            if os.path.exists(para_name):
                pass
            else:
                para, best = model_set(X, Y, split_type=split_type, model_type=model_type, model_dir=model_dir,
                                           file_name=file_name, FP_type=FP_type)  #

                print(best_name)
                para.to_csv(para_name, index=False)
                best.to_csv(best_name, index=False)
    return

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--file', required=True, help='we must give this para')
    parser.add_argument('--split', default=['scaffold'], nargs='*',choices = ['random','scaffold','cluster'])
    parser.add_argument('--FP', default=['MACCS'], nargs='*',choices = ['ECFP4','MACCS','2d-3d','pubchem'])
    parser.add_argument('--model', default=['SVM'], nargs='*',choices = ['DNN','KNN','SVM','RF','XGB'])
    parser.add_argument('--difftasks', default=['activity'], type=str)
    parser.add_argument('--threads', default=10,type=int)
    parser.add_argument('--mpl', default=False, type=str)

    args = parser.parse_args()
    return args

def main(file,split,FP,model,cpus,mpl,difftasks):
    tasks_dic = {'freesolv': ['activity'], 'esol': ['activity'], 'lipop': ['activity'], 'bace': ['activity'],
                 'bbbp': ['activity'], 'hiv': ['activity'],
                 'clintox': ['FDA_APPROVED', 'CT_TOX'],
                 'sider': ['SIDER1', 'SIDER2', 'SIDER3', 'SIDER4', 'SIDER5', 'SIDER6', 'SIDER7', 'SIDER8', 'SIDER9',
                           'SIDER10', 'SIDER11', 'SIDER12', 'SIDER13', 'SIDER14', 'SIDER15', 'SIDER16', 'SIDER17',
                           'SIDER18', 'SIDER19', 'SIDER20', 'SIDER21', 'SIDER22', 'SIDER23', 'SIDER24', 'SIDER25',
                           'SIDER26', 'SIDER27'],
                 'tox21': ['NR-AR', 'NR-AR-LBD', 'NR-AhR', 'NR-Aromatase', 'NR-ER', 'NR-ER-LBD', 'NR-PPAR-gamma', 'SR-ARE',
                           'SR-HSE', 'SR-MMP', 'SR-p53'],#'SR-ATAD5',
                 'muv': [
                     "MUV-466", "MUV-548", "MUV-600", "MUV-644", "MUV-652", "MUV-689", "MUV-692", "MUV-712", "MUV-713",
                     "MUV-733", "MUV-737", "MUV-810", "MUV-832", "MUV-846", "MUV-852", "MUV-858", "MUV-859"
                 ],
                 'toxcast': ['ACEA_T47D_80hr_Negative', 'ACEA_T47D_80hr_Positive',
                             'APR_HepG2_CellCycleArrest_24h_dn',
                             'APR_HepG2_CellCycleArrest_72h_dn', 'APR_HepG2_CellLoss_24h_dn',
                             'APR_HepG2_CellLoss_72h_dn', 'APR_HepG2_MicrotubuleCSK_72h_up',
                             'APR_HepG2_MitoMass_24h_dn', 'APR_HepG2_MitoMass_72h_dn',
                             'APR_HepG2_MitoMembPot_24h_dn', 'APR_HepG2_MitoMembPot_72h_dn',
                             'APR_HepG2_MitoticArrest_24h_up', 'APR_HepG2_MitoticArrest_72h_up',
                             'APR_HepG2_OxidativeStress_24h_up',
                             'APR_HepG2_OxidativeStress_72h_up',
                             'APR_HepG2_StressKinase_72h_up', 'APR_HepG2_p53Act_24h_up',
                             'APR_HepG2_p53Act_72h_up', 'ATG_AP_1_CIS_up', 'ATG_Ahr_CIS_up',
                             'ATG_BRE_CIS_up', 'ATG_CMV_CIS_up', 'ATG_CRE_CIS_up',
                             'ATG_DR4_LXR_CIS_dn', 'ATG_DR5_CIS_up', 'ATG_EGR_CIS_up',
                             'ATG_ERE_CIS_up', 'ATG_ERa_TRANS_up', 'ATG_E_Box_CIS_dn',
                             'ATG_HIF1a_CIS_up', 'ATG_HSE_CIS_up', 'ATG_IR1_CIS_dn',
                             'ATG_ISRE_CIS_dn', 'ATG_MRE_CIS_up', 'ATG_NRF2_ARE_CIS_up',
                             'ATG_Oct_MLP_CIS_up', 'ATG_PBREM_CIS_up', 'ATG_PPARg_TRANS_up',
                             'ATG_PPRE_CIS_up', 'ATG_PXRE_CIS_dn', 'ATG_PXRE_CIS_up',
                             'ATG_PXR_TRANS_up', 'ATG_Pax6_CIS_up', 'ATG_RORE_CIS_up',
                             'ATG_RXRb_TRANS_up', 'ATG_SREBP_CIS_up', 'ATG_Sp1_CIS_up',
                             'ATG_TCF_b_cat_CIS_dn', 'ATG_VDRE_CIS_up', 'ATG_Xbp1_CIS_up',
                             'ATG_p53_CIS_dn', 'BSK_3C_Eselectin_down', 'BSK_3C_HLADR_down',
                             'BSK_3C_ICAM1_down', 'BSK_3C_IL8_down', 'BSK_3C_MCP1_down',
                             'BSK_3C_MIG_down', 'BSK_3C_Proliferation_down', 'BSK_3C_SRB_down',
                             'BSK_3C_Thrombomodulin_up', 'BSK_3C_TissueFactor_down',
                             'BSK_3C_VCAM1_down', 'BSK_3C_Vis_down', 'BSK_3C_uPAR_down',
                             'BSK_4H_Eotaxin3_down', 'BSK_4H_MCP1_down',
                             'BSK_4H_Pselectin_down', 'BSK_4H_SRB_down', 'BSK_4H_VCAM1_down',
                             'BSK_4H_VEGFRII_down', 'BSK_4H_uPAR_down', 'BSK_BE3C_HLADR_down',
                             'BSK_BE3C_IL1a_down', 'BSK_BE3C_IP10_down', 'BSK_BE3C_MIG_down',
                             'BSK_BE3C_MMP1_down', 'BSK_BE3C_MMP1_up', 'BSK_BE3C_PAI1_down',
                             'BSK_BE3C_SRB_down', 'BSK_BE3C_TGFb1_down', 'BSK_BE3C_tPA_down',
                             'BSK_BE3C_uPAR_down', 'BSK_BE3C_uPA_down', 'BSK_CASM3C_HLADR_down',
                             'BSK_CASM3C_IL6_down', 'BSK_CASM3C_IL8_down',
                             'BSK_CASM3C_LDLR_down', 'BSK_CASM3C_MCP1_down',
                             'BSK_CASM3C_MCSF_down', 'BSK_CASM3C_MIG_down',
                             'BSK_CASM3C_Proliferation_down', 'BSK_CASM3C_SAA_down',
                             'BSK_CASM3C_SRB_down', 'BSK_CASM3C_Thrombomodulin_up',
                             'BSK_CASM3C_TissueFactor_down', 'BSK_CASM3C_VCAM1_down',
                             'BSK_CASM3C_uPAR_down', 'BSK_KF3CT_ICAM1_down',
                             'BSK_KF3CT_IL1a_down', 'BSK_KF3CT_IP10_down',
                             'BSK_KF3CT_MCP1_down', 'BSK_KF3CT_MMP9_down', 'BSK_KF3CT_SRB_down',
                             'BSK_KF3CT_TGFb1_down', 'BSK_KF3CT_TIMP2_down',
                             'BSK_KF3CT_uPA_down', 'BSK_LPS_CD40_down',
                             'BSK_LPS_Eselectin_down', 'BSK_LPS_IL1a_down', 'BSK_LPS_IL8_down',
                             'BSK_LPS_MCP1_down', 'BSK_LPS_MCSF_down', 'BSK_LPS_PGE2_down',
                             'BSK_LPS_SRB_down', 'BSK_LPS_TNFa_down',
                             'BSK_LPS_TissueFactor_down', 'BSK_LPS_VCAM1_down',
                             'BSK_SAg_CD38_down', 'BSK_SAg_CD40_down', 'BSK_SAg_CD69_down',
                             'BSK_SAg_Eselectin_down', 'BSK_SAg_IL8_down', 'BSK_SAg_MCP1_down',
                             'BSK_SAg_MIG_down', 'BSK_SAg_PBMCCytotoxicity_down',
                             'BSK_SAg_Proliferation_down', 'BSK_SAg_SRB_down',
                             'BSK_hDFCGF_CollagenIII_down', 'BSK_hDFCGF_IL8_down',
                             'BSK_hDFCGF_IP10_down', 'BSK_hDFCGF_MCSF_down',
                             'BSK_hDFCGF_MIG_down', 'BSK_hDFCGF_MMP1_down',
                             'BSK_hDFCGF_PAI1_down', 'BSK_hDFCGF_Proliferation_down',
                             'BSK_hDFCGF_SRB_down', 'BSK_hDFCGF_TIMP1_down',
                             'BSK_hDFCGF_VCAM1_down', 'CEETOX_H295R_11DCORT_dn',
                             'CEETOX_H295R_ANDR_dn', 'CEETOX_H295R_CORTISOL_dn',
                             'CEETOX_H295R_ESTRONE_dn', 'CEETOX_H295R_ESTRONE_up',
                             'NHEERL_ZF_144hpf_TERATOSCORE_up', 'NVS_NR_bER', 'NVS_NR_hER',
                             'NVS_NR_hPPARg', 'NVS_NR_hPXR', 'NVS_NR_mERa', 'OT_AR_ARSRC1_0960',
                             'OT_ER_ERaERb_0480', 'OT_ER_ERaERb_1440', 'OT_ER_ERbERb_0480',
                             'OT_ER_ERbERb_1440', 'OT_ERa_EREGFP_0120', 'OT_FXR_FXRSRC1_0480',
                             'OT_NURR1_NURR1RXRa_0480', 'TOX21_ARE_BLA_agonist_ratio',
                             'TOX21_AR_BLA_Antagonist_ratio', 'TOX21_AR_LUC_MDAKB2_Antagonist',
                             'TOX21_AR_LUC_MDAKB2_Antagonist2', 'TOX21_AhR_LUC_Agonist',
                             'TOX21_Aromatase_Inhibition', 'TOX21_ERa_BLA_Antagonist_ratio',
                             'TOX21_ERa_LUC_BG1_Agonist', 'TOX21_FXR_BLA_antagonist_ratio',
                             'TOX21_MMP_ratio_down', 'TOX21_TR_LUC_GH3_Antagonist',
                             'TOX21_p53_BLA_p1_ratio', 'TOX21_p53_BLA_p2_ch2',
                             'TOX21_p53_BLA_p2_ratio', 'TOX21_p53_BLA_p2_viability',
                             'TOX21_p53_BLA_p3_ratio', 'TOX21_p53_BLA_p4_ratio',
                             'TOX21_p53_BLA_p5_ratio', 'Tanguay_ZF_120hpf_AXIS_up',
                             'Tanguay_ZF_120hpf_ActivityScore', 'Tanguay_ZF_120hpf_JAW_up',
                             'Tanguay_ZF_120hpf_MORT_up', 'Tanguay_ZF_120hpf_PE_up',
                             'Tanguay_ZF_120hpf_SNOU_up', 'Tanguay_ZF_120hpf_YSE_up']}
    difftasks = tasks_dic[difftasks] if not difftasks ==["activity"] else ['activity']

    data = pd.read_csv(file,error_bad_lines=False)
    dic = {'split':split,'FP':FP,'model':model}
    hy = pd.DataFrame(list(itertools.product(*dic.values())),columns=dic.keys())

    if mpl == False:
        for i in range(len(hy.index.values)):
            start_time = time.time()
            split_type = hy.iloc[i].split
            FP_type = hy.iloc[i].FP
            model_type = hy.iloc[i].model
            a = pair_param(file, data, split_type, model_type, FP_type,difftasks)
            use_time = round(time.time() - start_time,5)


            local_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
            record_save = file.replace(file.split('dataset')[-1],'/record_every_model.csv')
            f = open(record_save, 'a+')
            f.write(",".join([file, split_type, model_type, str(use_time),local_time,FP_type,'\n']))
            f.close()
    else:
        p = mp.Pool(processes = cpus)
        for i in range(len(hy.index.values)):
            start_time = time.time()
            split_type = hy.iloc[i].split
            FP_type = hy.iloc[i].FP
            model_type = hy.iloc[i].model
            # print(i)
            result = p.apply_async(pair_param,args=(file,data,split_type,model_type,FP_type,difftasks))
            use_time = int(time.time() - start_time)
            # if use_time >= 10:
            local_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
            record_save = file.replace(file.split('dataset')[-1], '/record_every_model.csv')
            f = open(record_save, 'a+')
            f.write(",".join([file, split_type, model_type, str(use_time),local_time,FP_type,'\n']))
            f.close()
        p.close()
        p.join()

if __name__ == '__main__':
    args = parse_args()
    file = args.file
    split = args.split
    FP = args.FP
    model = args.model
    cpus=args.threads
    mpl = args.mpl
    difftasks = args.difftasks
    main(file,split,FP,model,cpus,mpl,difftasks)